/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/base64/base64.c                                          *
 * Created:     2004-10-01 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include "base64.h"


#ifdef MISC_BUF_SIZE
#define MISC_BASE64_BUFSIZE MISC_BUF_SIZE
#else
#define MISC_BASE64_BUFSIZE 4096
#endif


static const char *argv0 = NULL;

static unsigned par_encode = 1;
static unsigned par_base = 64;
static unsigned par_cols = 76;
static int      par_preserve = 0;
static int      par_delim = 1;
static int      par_compress_y = 0;
static int      par_compress_z = 0;


static char base32_enc[32] = {
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
	'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
	'Y', 'Z', '2', '3', '4', '5', '6', '7'
};

static unsigned char base32_dec[96] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* 20 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, /* 30 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, /* 40 */
	0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
	0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, /* 50 */
	0x17, 0x18, 0x19, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, /* 60 */
	0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
	0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, /* 70 */
	0x17, 0x18, 0x19, 0xff, 0xff, 0xff, 0xff, 0xff,
};


static char base64_enc[64] = {
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
	'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
	'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
	'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
	'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
	'w', 'x', 'y', 'z', '0', '1', '2', '3',
	'4', '5', '6', '7', '8', '9', '+', '/'
};

static unsigned char base64_dec[96] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* 20 */
	0xff, 0xff, 0xff, 0x3e, 0xff, 0xff, 0xff, 0x3f,
	0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, /* 30 */
	0x3c, 0x3d, 0xff, 0xff, 0xff, 0x40, 0xff, 0xff,
	0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, /* 40 */
	0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
	0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, /* 50 */
	0x17, 0x18, 0x19, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, /* 60 */
	0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
	0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, /* 70 */
	0x31, 0x32, 0x33, 0xff, 0xff, 0xff, 0xff, 0xff,
};


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'b', 1, "base", "int", "Set the base [64]" },
	{ 'c', 1, "columns", "int", "Set number of bytes per row [76]" },
	{ 'd', 0, "decode", NULL, "Decode text to binary" },
	{ 'e', 0, "encode", NULL, "Encode binary to text [default]" },
	{ 'i', 1, "input", "string", "Set input file name [stdin]" },
	{ 'l', 0, "no-delimiter", NULL, "Don't use a delimiter for base85 [no]" },
	{ 'o', 1, "output", "string", "Set output file name [stdout]" },
	{ 'p', 0, "preserve", NULL, "Preserve characters with base85 [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'y', 0, "compress-y", NULL, "Use text compression with base85 [no]" },
	{ 'z', 0, "compress-z", NULL, "Use zero compression with base85 [no]" },
	{  -1, 0, NULL, NULL, NULL }
};

static
void print_help (void)
{
	misc_getopt_help (
		"base64: base16, base32, base64 or base85 conversion",
		"usage: base64 [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

/*
 * Print a character and keep track of the current column.
 */
static
void base_fputc (FILE *fp, int c, unsigned *col)
{
	fputc (c, fp);

	*col += 1;

	if (par_cols > 0) {
		if (*col >= par_cols) {
			fputc ('\n', fp);
			*col = 0;
		}
	}
}


static
void b16_encode_group (FILE *fp, unsigned val, unsigned *col)
{
	unsigned dig;

	dig = (val >> 4) & 0x0f;
	dig = (dig <= 9) ? (dig + '0') : (dig + 'A' - 10);
	base_fputc (fp, dig, col);

	dig = val & 0x0f;
	dig = (dig <= 9) ? (dig + '0') : (dig + 'A' - 10);
	base_fputc (fp, dig, col);
}

static
int b16_encode (FILE *inp, FILE *out)
{
	unsigned      i, n;
	unsigned      col;
	unsigned char buf[MISC_BASE64_BUFSIZE];

	col = 0;

	while (1) {
		n = fread (buf, 1, MISC_BASE64_BUFSIZE, inp);

		if (n == 0) {
			break;
		}

		for (i = 0; i < n; i++) {
			b16_encode_group (out, buf[i], &col);
		}
	}

	if ((par_cols > 0) && (col > 0)) {
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

static
int b16_decode (FILE *inp, FILE *out)
{
	int           c;
	unsigned      dig, cnt;
	unsigned long val;

	val = 0;
	cnt = 0;

	while (1) {
		c = fgetc (inp);

		if (c == EOF) {
			break;
		}
		else if ((c >= '0') && (c <= '9')) {
			dig = c - '0';
		}
		else if ((c >= 'A') && (c <= 'F')) {
			dig = c - 'A' + 10;
		}
		else if ((c >= 'a') && (c <= 'f')) {
			dig = c - 'a' + 10;
		}
		else if ((c <= 0x20) || (c == 0x7f)) {
			continue;
		}
		else {
			fprintf (stderr, "%s: invalid character (0x%02x)\n",
				argv0, c
			);

			return (1);
		}

		val = (val << 4) | dig;
		cnt += 1;

		if (cnt >= 2) {
			fputc (val & 0xff, out);
			cnt = 0;
		}
	}

	if (cnt > 0) {
		fprintf (stderr, "%s: truncated input\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}


static
void b32_fputc (FILE *fp, unsigned v, unsigned *col)
{
	base_fputc (fp, base32_enc[v & 0x1f], col);
}

static
void b32_encode_group (FILE *fp, const unsigned char *buf, unsigned cnt, unsigned *col)
{
	b32_fputc (fp, buf[0] >> 3, col);
	b32_fputc (fp, (buf[0] << 2) | (buf[1] >> 6), col);

	if (cnt > 1) {
		b32_fputc (fp, buf[1] >> 1, col);
		b32_fputc (fp, (buf[1] << 4) | (buf[2] >> 4), col);
	}
	else {
		base_fputc (fp, '=', col);
		base_fputc (fp, '=', col);
	}

	if (cnt > 2) {
		b32_fputc (fp, (buf[2] << 1) | (buf[3] >> 7), col);
	}
	else {
		base_fputc (fp, '=', col);
	}

	if (cnt > 3) {
		b32_fputc (fp, buf[3] >> 2, col);
		b32_fputc (fp, (buf[3] << 3) | (buf[4] >> 5), col);
	}
	else {
		base_fputc (fp, '=', col);
		base_fputc (fp, '=', col);
	}

	if (cnt > 4) {
		b32_fputc (fp, buf[4], col);
	}
	else {
		base_fputc (fp, '=', col);
	}
}

static
int b32_encode (FILE *inp, FILE *out)
{
	unsigned      i, j, n;
	unsigned      col;
	unsigned char buf[MISC_BASE64_BUFSIZE + 8];

	col = 0;

	n = fread (buf, 1, MISC_BASE64_BUFSIZE, inp);

	while (n >= 5) {
		i = 0;
		j = 0;

		while ((i + 4) < n) {
			b32_encode_group (out, buf + i, 5, &col);
			i += 5;
		}

		while (i < n) {
			buf[j++] = buf[i++];
		}

		n = j + fread (buf + j, 1, MISC_BASE64_BUFSIZE, inp);
	}

	if (n > 0) {
		for (i = n; i < 5; i++) {
			buf[i] = 0;
		}

		b32_encode_group (out, buf, n, &col);
	}

	if ((par_cols > 0) && (col > 0)) {
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

static
int b32_decode (FILE *inp, FILE *out)
{
	int           c;
	unsigned      dig, bit;
	unsigned long val;

	val = 0;
	bit = 0;

	while (1) {
		c = fgetc (inp);

		if ((c == EOF) || (c == '=')) {
			break;
		}
		else if ((c >= 32) && (c < 127)) {
			dig = base32_dec[c - 32];
		}
		else if ((c < 32) || (c == 127)) {
			continue;
		}
		else {
			dig = 255;
		}

		if (dig > 0x1f) {
			fprintf (stderr, "%s: invalid character (0x%02x)\n",
				argv0, c
			);
			return (1);
		}

		val = (val << 5) | dig;
		bit += 5;

		while (bit >= 8) {
			fputc ((val >> (bit - 8)) & 0xff, out);
			bit -= 8;
		}
	}

	fflush (out);

	return (0);
}


static
void b64_fputc (FILE *fp, unsigned v, unsigned *col)
{
	base_fputc (fp, base64_enc[v & 0x3f], col);
}

static
void b64_encode_group (FILE *fp, const unsigned char *buf, unsigned cnt, unsigned *col)
{
	b64_fputc (fp, buf[0] >> 2, col);
	b64_fputc (fp, (buf[0] << 4) | (buf[1] >> 4), col);

	if (cnt > 1) {
		b64_fputc (fp, (buf[1] << 2) | (buf[2] >> 6), col);
	}
	else {
		base_fputc (fp, '=', col);
	}

	if (cnt > 2) {
		b64_fputc (fp, buf[2], col);
	}
	else {
		base_fputc (fp, '=', col);
	}
}

static
int b64_encode (FILE *inp, FILE *out)
{
	unsigned      i, j, n;
	unsigned      col;
	unsigned char buf[MISC_BASE64_BUFSIZE + 4];

	col = 0;

	n = fread (buf, 1, MISC_BASE64_BUFSIZE, inp);

	while (n >= 3) {
		i = 0;
		j = 0;

		while ((i + 2) < n) {
			b64_encode_group (out, buf + i, 3, &col);
			i += 3;
		}

		while (i < n) {
			buf[j++] = buf[i++];
		}

		n = j + fread (buf + j, 1, MISC_BASE64_BUFSIZE, inp);
	}

	if (n > 0) {
		for (i = n; i < 3; i++) {
			buf[i] = 0;
		}

		b64_encode_group (out, buf, n, &col);
	}

	if ((par_cols > 0) && (col > 0)) {
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

static
int b64_decode (FILE *inp, FILE *out)
{
	int           c;
	unsigned      dig, bit;
	unsigned long val;

	val = 0;
	bit = 0;

	while (1) {
		c = fgetc (inp);

		if ((c == EOF) || (c == '=')) {
			break;
		}
		else if ((c >= 32) && (c < 127)) {
			dig = base64_dec[c - 32];
		}
		else if ((c < 32) || (c == 127)) {
			continue;
		}
		else {
			dig = 255;
		}

		if (dig > 0x3f) {
			fprintf (stderr, "%s: invalid character (0x%02x)\n",
				argv0, c
			);
			return (1);
		}

		val = (val << 6) | dig;
		bit += 6;

		while (bit >= 8) {
			fputc ((val >> (bit - 8)) & 0xff, out);
			bit -= 8;
		}
	}

	fflush (out);

	return (0);
}


static
void b85_fputc (FILE *fp, unsigned char v, unsigned *col)
{
	base_fputc (fp, v + 33, col);
}

static
void b85_encode_group (FILE *fp, const unsigned char *buf, unsigned cnt, unsigned *col)
{
	unsigned      i;
	unsigned long v;
	unsigned char out[5];

	if ((cnt == 0) || (cnt > 4)) {
		return;
	}

	v = buf[0];
	v = (v << 8) | ((cnt > 1) ? buf[1] : 0);
	v = (v << 8) | ((cnt > 2) ? buf[2] : 0);
	v = (v << 8) | ((cnt > 3) ? buf[3] : 0);

	if (par_compress_z) {
		if (cnt == 4) {
			if (v == 0) {
				base_fputc (fp, 'z', col);
				return;
			}
		}
	}

	if (par_compress_y) {
		if (cnt == 4) {
			if (v == 0x20202020) {
				base_fputc (fp, 'y', col);
				return;
			}
		}
	}

	cnt += 1;

	for (i = 0; i < 5; i++) {
		out[4 - i] = v % 85;
		v = v / 85;
	}

	for (i = 0; i < cnt; i++) {
		b85_fputc (fp, out[i], col);
	}
}

static
int b85_encode (FILE *inp, FILE *out)
{
	unsigned      i, j, n;
	unsigned      col;
	unsigned char buf[MISC_BASE64_BUFSIZE + 4];

	col = 0;

	if (par_delim) {
		fputs ("<~", out);
		col = 2;
	}

	n = fread (buf, 1, MISC_BASE64_BUFSIZE, inp);

	while (n >= 4) {
		i = 0;
		j = 0;

		while ((i + 3) < n) {
			b85_encode_group (out, buf + i, 4, &col);
			i += 4;
		}

		while (i < n) {
			buf[j++] = buf[i++];
		}

		n = j + fread (buf + j, 1, MISC_BASE64_BUFSIZE, inp);
	}

	if (n > 0) {
		b85_encode_group (out, buf, n, &col);
	}

	if (par_delim) {
		fputs ("~>", out);
	}

	if ((par_cols > 0) && (col > 0)) {
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

static
void b85_fputw (FILE *fp, unsigned long val)
{
	fputc ((val >> 24) & 0xff, fp);
	fputc ((val >> 16) & 0xff, fp);
	fputc ((val >> 8) & 0xff, fp);
	fputc (val & 0xff, fp);
}

/*
 * Scan the input stream for the starting delimiter.
 */
static
int b85_seek_start (FILE *inp, FILE *out)
{
	int c1, c2;

	while (1) {
		c1 = fgetc (inp);

		if (c1 == EOF) {
			return (1);
		}

		if (c1 == '<') {
			c2 = fgetc (inp);
			if (c2 == '~') {
				return (0);
			}

			ungetc (c2, inp);
		}

		if (par_preserve) {
			fputc (c1, out);
		}
	}

	return (1);
}

static
int b85_decode_run (FILE *inp, FILE *out)
{
	int           c;
	unsigned      i, cnt;
	unsigned      dig;
	unsigned long val, fct;

	cnt = 0;
	val = 0;
	fct = 1;

	while (1) {
		c = fgetc (inp);
		if (c == EOF) {
			break;
		}

		if ((c >= 33) && (c <= 117)) {
			dig = c - 33;

			val = 85 * val + dig;
			cnt += 1;

			if (cnt >= 5) {
				b85_fputw (out, val);

				cnt = 0;
				val = 0;
			}
		}
		else if ((c == 'y') && (cnt == 0)) {
			b85_fputw (out, 0x20202020);
		}
		else if ((c == 'z') && (cnt == 0)) {
			b85_fputw (out, 0x00000000);
		}
		else if ((par_delim == 0) && (c == 'x')) {
			break;
		}
		else if ((par_delim != 0) && (c == '~')) {
			if (fgetc (inp) == '>') {
				break;
			}

			fprintf (stderr, "%s: invalid character (0x%02x)\n",
				argv0, c
			);

			return (1);
		}
		else if ((c <= 32) || (c == 127)) {
			continue;
		}
		else {
			fprintf (stderr, "%s: invalid character (0x%02x)\n",
				argv0, c
			);
			return (1);
		}
	}

	if (cnt == 1) {
		fprintf (stderr, "%s: truncated input\n", argv0);
		return (1);
	}
	else if (cnt > 1) {
		for (i = cnt; i < 5; i++) {
			val = 85 * val + 84;
		}

		for (i = 1; i < cnt; i++) {
			fputc ((val >> 24) & 0xff, out);
			val = val << 8;
		}
	}

	fflush (out);

	return (0);
}

static
int b85_decode (FILE *inp, FILE *out)
{
	while (1) {
		if (par_delim) {
			if (b85_seek_start (inp, out)) {
				return (0);
			}
		}

		if (b85_decode_run (inp, out)) {
			return (1);
		}

		if (par_delim == 0) {
			return (0);
		}
	}

	return (1);
}

static
int str_tail (const char *str, const char *tail)
{
	unsigned n1, n2;

	n1 = strlen (str);
	n2 = strlen (tail);

	str += n1;
	tail += n2;

	if (n2 > n1) {
		return (0);
	}

	while (n2 > 0) {
		str -= 1;
		tail -= 1;

		if (*str != *tail) {
			return (0);
		}

		n2 -= 1;
	}

	return (1);
}

static
void set_default_base (const char *argv0)
{
	if (str_tail (argv0, "base16")) {
		par_base = 16;
	}
	else if (str_tail (argv0, "base32")) {
		par_base = 32;
	}
	else if (str_tail (argv0, "base64")) {
		par_base = 64;
	}
	else if (str_tail (argv0, "base85")) {
		par_base = 85;
	}
	else {
		par_base = 64;
	}
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	set_default_base (argv0);

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("base64");
			return (EXIT_SUCCESS);

		case 'b':
			if (str_get_uint (optarg[0], &par_base)) {
				fprintf (stderr, "%s: bad base (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'c':
			if (str_get_uint (optarg[0], &par_cols)) {
				fprintf (stderr, "%s: bad column count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'd':
			par_encode = 0;
			break;

		case 'e':
			par_encode = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'l':
			par_delim = 0;
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			par_preserve = 1;
			break;

		case 'y':
			par_compress_y = 1;
			break;

		case 'z':
			par_compress_z = 1;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, par_encode ? "rb" : "r");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, par_encode ? "w" : "wb");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	if (par_encode) {
		if (par_base == 85) {
			r = b85_encode (inp, out);
		}
		else if (par_base == 64) {
			r = b64_encode (inp, out);
		}
		else if (par_base == 32) {
			r = b32_encode (inp, out);
		}
		else if (par_base == 16) {
			r = b16_encode (inp, out);
		}
		else {
			fprintf (stderr, "%s: unsupported base (%u)\n",
				argv0, par_base
			);
			r = 1;
		}
	}
	else {
		if (par_base == 85) {
			r = b85_decode (inp, out);
		}
		else if (par_base == 64) {
			r = b64_decode (inp, out);
		}
		else if (par_base == 32) {
			r = b32_decode (inp, out);
		}
		else if (par_base == 16) {
			r = b16_decode (inp, out);
		}
		else {
			fprintf (stderr, "%s: unsupported base (%u)\n",
				argv0, par_base
			);
			r = 1;
		}
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (EXIT_FAILURE);
	}

	if (ferror (out)) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (EXIT_FAILURE);
	}

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
