/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/file.h                                               *
 * Created:     2002-02-05 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#ifndef MISC_FILE_H
#define MISC_FILE_H 1


#define MISC_FILE_BUF_SIZE 4096

#define MISC_MODE_NONE 0
#define MISC_MODE_INP  1
#define MISC_MODE_OUT  2
#define MISC_MODE_APP  3


typedef struct {
	int           mode;

	char          *name;
	int           fd;
	int           close;

	unsigned long buf_i;
	unsigned long buf_n;
	unsigned char buf[MISC_FILE_BUF_SIZE];
} file_t;


void file_init (file_t *f);
void file_set_name (file_t *f, const char *str);
void file_set_fd (file_t *f, int fd, int close);
void file_set_stdout (file_t *f);
void file_set_stdin (file_t *f);

int file_open_out (file_t *f);
int file_open_app (file_t *f);
int file_open_inp (file_t *f);

int file_flush (file_t *f);
void file_close (file_t *f);

unsigned long file_write (file_t *f, const void *buf, unsigned long n);
unsigned long file_write_block (file_t *f, const void *buf, unsigned long n);
unsigned long file_write_string (file_t *f, const char *str);
unsigned long file_read (file_t *f, void *buf, unsigned long n);
unsigned long file_read_block (file_t *f, void *buf, unsigned long n);
unsigned long file_skip_block (file_t *f, unsigned long n);

int file_get_char (file_t *f);
int file_unget_char (file_t *f, int c);
int file_set_char (file_t *f, int c);
int file_get_byte (file_t *f, unsigned char *c);
int file_set_byte (file_t *f, unsigned char c);
int file_unget_byte (file_t *f, unsigned char c);


#endif
