/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/md2.h                                                *
 * Created:     2008-08-10 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2008-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#ifndef MISC_MD2_H
#define MISC_MD2_H 1


/*!***************************************************************************
 * @short The MD2 context
 *****************************************************************************/
typedef struct {
	unsigned char state[48];
	unsigned char chksum[16];

	unsigned char buf[16];
	unsigned      buf_i;
} md2_t;



/*!***************************************************************************
 * @short Initialize an MD2 context
 *****************************************************************************/
void md2_init (md2_t *md2);

/*!***************************************************************************
 * @short Copy an MD2 context
 *****************************************************************************/
void md2_copy (md2_t *dst, const md2_t *src);

/*!***************************************************************************
 * @short Calculate an MD2 message digest
 * @param md2 The MD2 context
 * @param buf The data
 * @param cnt The number of bytes at buf
 *
 * This function can be called multiple times to add more bytes.
 *****************************************************************************/
void md2_calc (md2_t *md2, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short End calculating an MD2 message digest
 *
 * This function must be called after the last call to md2_calc() and
 * before getting the md2 message digest.
 *****************************************************************************/
void md2_done (md2_t *md2);

/*!***************************************************************************
 * @short  Copy the MD2 message digest to a byte buffer
 * @param  md2 The MD2 context
 * @param  buf The MD2 buffer
 * @param  max The buffer size
 * @return The number of bytes written to the buffer
 *****************************************************************************/
unsigned md2_get_digest (md2_t *md2, void *buf, unsigned max);

/*!***************************************************************************
 * @short Calculate the MD2 message digest of a buffer
 * @param dst The 16 byte MD2 buffer
 * @param src The source buffer
 * @param n   The size in bytes of the source buffer
 *****************************************************************************/
void md2_calc1 (void *dst, const void *src, unsigned long n);


#endif
