/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/open.c                                               *
 * Created:     2009-10-22 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2009 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/open.h>


#ifdef MISC_DOS

#include <io.h>
#include <fcntl.h>

static
int misc_is_binary (const char *mode)
{
	while (*mode != 0) {
		if ((*mode == 'b') || (*mode == 'B')) {
			return (1);
		}

		mode += 1;
	}

	return (0);
}

static
void misc_set_mode (FILE *fp, const char *mode)
{
	int fd, dd, nd;

	fd = fileno (fp);

	if (misc_is_binary (mode)) {
		setmode (fd, O_BINARY);
		fp->flags |= _F_BIN;
		nd = 0x20;
	}
	else {
		setmode (fd, O_TEXT);
		fp->flags &= ~_F_BIN;
		nd = 0x00;
	}

	dd = ioctl (fd, 0, (void *) 0, 0);

	if ((dd != -1) && (dd & 0x80)) {
		if ((dd & 0x20) != nd) {
			dd = (dd & 0xdf) | nd;
			ioctl (fd, 1, (void *) dd, 0);
		}
	}
}

#else

#define misc_set_mode(fp, mode)

#endif


FILE *misc_open_inp (const char *argv0, const char *name, const char *mode)
{
	FILE *fp;

	if ((name == NULL) || (strcmp (name, "-") == 0)) {
		fp = stdin;
	}
	else {
		fp = fopen (name, mode);
	}

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n", argv0, name);
	}

	misc_set_mode (fp, mode);

	return (fp);
}

FILE *misc_open_out (const char *argv0, const char *name, const char *mode)
{
	FILE *fp;

	if ((name == NULL) || (strcmp (name, "-") == 0)) {
		fp = stdout;
	}
	else if (strcmp (name, "--") == 0) {
		fp = stderr;
	}
	else {
		fp = fopen (name, mode);
	}

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n", argv0, name);
	}

	misc_set_mode (fp, mode);

	return (fp);
}

void misc_close (FILE *fp)
{
	if (fp == NULL) {
		return;
	}

	if ((fp == stdin) || (fp == stdout) || (fp == stderr)) {
		return;
	}

	fclose (fp);
}
