/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/sha512.c                                             *
 * Created:     2005-11-18 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/config.h>
#include <lib/sha512.h>


static
unsigned long long sha512_H0[8] = {
	0x6a09e667f3bcc908ULL,
	0xbb67ae8584caa73bULL,
	0x3c6ef372fe94f82bULL,
	0xa54ff53a5f1d36f1ULL,
	0x510e527fade682d1ULL,
	0x9b05688c2b3e6c1fULL,
	0x1f83d9abfb41bd6bULL,
	0x5be0cd19137e2179ULL
};

static
unsigned long long sha512_K[80] = {
	0x428a2f98d728ae22ULL, 0x7137449123ef65cdULL,
	0xb5c0fbcfec4d3b2fULL, 0xe9b5dba58189dbbcULL,
	0x3956c25bf348b538ULL, 0x59f111f1b605d019ULL,
	0x923f82a4af194f9bULL, 0xab1c5ed5da6d8118ULL,
	0xd807aa98a3030242ULL, 0x12835b0145706fbeULL,
	0x243185be4ee4b28cULL, 0x550c7dc3d5ffb4e2ULL,
	0x72be5d74f27b896fULL, 0x80deb1fe3b1696b1ULL,
	0x9bdc06a725c71235ULL, 0xc19bf174cf692694ULL,
	0xe49b69c19ef14ad2ULL, 0xefbe4786384f25e3ULL,
	0x0fc19dc68b8cd5b5ULL, 0x240ca1cc77ac9c65ULL,
	0x2de92c6f592b0275ULL, 0x4a7484aa6ea6e483ULL,
	0x5cb0a9dcbd41fbd4ULL, 0x76f988da831153b5ULL,
	0x983e5152ee66dfabULL, 0xa831c66d2db43210ULL,
	0xb00327c898fb213fULL, 0xbf597fc7beef0ee4ULL,
	0xc6e00bf33da88fc2ULL, 0xd5a79147930aa725ULL,
	0x06ca6351e003826fULL, 0x142929670a0e6e70ULL,
	0x27b70a8546d22ffcULL, 0x2e1b21385c26c926ULL,
	0x4d2c6dfc5ac42aedULL, 0x53380d139d95b3dfULL,
	0x650a73548baf63deULL, 0x766a0abb3c77b2a8ULL,
	0x81c2c92e47edaee6ULL, 0x92722c851482353bULL,
	0xa2bfe8a14cf10364ULL, 0xa81a664bbc423001ULL,
	0xc24b8b70d0f89791ULL, 0xc76c51a30654be30ULL,
	0xd192e819d6ef5218ULL, 0xd69906245565a910ULL,
	0xf40e35855771202aULL, 0x106aa07032bbd1b8ULL,
	0x19a4c116b8d2d0c8ULL, 0x1e376c085141ab53ULL,
	0x2748774cdf8eeb99ULL, 0x34b0bcb5e19b48a8ULL,
	0x391c0cb3c5c95a63ULL, 0x4ed8aa4ae3418acbULL,
	0x5b9cca4f7763e373ULL, 0x682e6ff3d6b2b8a3ULL,
	0x748f82ee5defb2fcULL, 0x78a5636f43172f60ULL,
	0x84c87814a1f0ab72ULL, 0x8cc702081a6439ecULL,
	0x90befffa23631e28ULL, 0xa4506cebde82bde9ULL,
	0xbef9a3f7b2c67915ULL, 0xc67178f2e372532bULL,
	0xca273eceea26619cULL, 0xd186b8c721c0c207ULL,
	0xeada7dd6cde0eb1eULL, 0xf57d4f7fee6ed178ULL,
	0x06f067aa72176fbaULL, 0x0a637dc5a2c898a6ULL,
	0x113f9804bef90daeULL, 0x1b710b35131c471bULL,
	0x28db77f523047d84ULL, 0x32caab7b40c72493ULL,
	0x3c9ebe0a15c9bebcULL, 0x431d67c49c100d4cULL,
	0x4cc5d4becb3e42b6ULL, 0x597f299cfc657e2aULL,
	0x5fcb6fab3ad6faecULL, 0x6c44198c4a475817ULL
};


static inline
unsigned long long sha512_add (unsigned long long x, unsigned long long y)
{
	return ((x + y) & 0xffffffffffffffffULL);
}

static inline
unsigned long long sha512_ror (unsigned long long x, unsigned n)
{
	x = (x >> n) | (x << (64 - n));

	return (x & 0xffffffffffffffffULL);
}

static inline
unsigned long long sha512_S0 (unsigned long long x)
{
	return (sha512_ror (x, 28) ^ sha512_ror (x, 34) ^ sha512_ror (x, 39));
}

static inline
unsigned long long sha512_S1 (unsigned long long x)
{
	return (sha512_ror (x, 14) ^ sha512_ror (x, 18) ^ sha512_ror (x, 41));
}

static inline
unsigned long long sha512_s0 (unsigned long long x)
{
	return (sha512_ror (x, 1) ^ sha512_ror (x, 8) ^ (x >> 7));
}

static inline
unsigned long long sha512_s1 (unsigned long long x)
{
	return (sha512_ror (x, 19) ^ sha512_ror (x, 61) ^ (x >> 6));
}

static inline
unsigned long long sha512_F (unsigned long long x, unsigned long long y, unsigned long long z)
{
	return (((x & y) ^ (~x & z)) & 0xffffffffffffffffULL);
}

static inline
unsigned long long sha512_G (unsigned long long x, unsigned long long y, unsigned long long z)
{
	return (((x & y) ^ (x & z) ^ (y & z)) & 0xffffffffffffffffULL);
}


void sha512_init (sha512_t *sha)
{
	unsigned i;

	sha->buf_i = 0;

	sha->cnt[0] = 0;
	sha->cnt[1] = 0;

	for (i = 0; i < 8; i++) {
		sha->H[i] = sha512_H0[i];
	}
}

void sha512_copy (sha512_t *dst, const sha512_t *src)
{
	unsigned i;

	dst->buf_i = src->buf_i;

	for (i = 0; i < src->buf_i; i++) {
		dst->buf[i] = src->buf[i];
	}

	dst->cnt[0] = src->cnt[0];
	dst->cnt[1] = src->cnt[1];

	for (i = 0; i < 8; i++) {
		dst->H[i] = src->H[i];
	}
}

static inline
void sha512_set_uint64 (void *buf, unsigned i, unsigned long long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (v >> 56) & 0xff;
	tmp[1] = (v >> 48) & 0xff;
	tmp[2] = (v >> 40) & 0xff;
	tmp[3] = (v >> 32) & 0xff;
	tmp[4] = (v >> 24) & 0xff;
	tmp[5] = (v >> 16) & 0xff;
	tmp[6] = (v >> 8) & 0xff;
	tmp[7] = v & 0xff;
}

static inline
unsigned long long sha512_get_uint64 (const void *buf, unsigned i)
{
	unsigned long long  v;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	v = tmp[0];
	v = (v << 8) | tmp[1];
	v = (v << 8) | tmp[2];
	v = (v << 8) | tmp[3];
	v = (v << 8) | tmp[4];
	v = (v << 8) | tmp[5];
	v = (v << 8) | tmp[6];
	v = (v << 8) | tmp[7];

	return (v);
}

static
void sha512_add_size (sha512_t *sha, unsigned long long n)
{
	n &= 0xffffffffffffffffULL;

	sha->cnt[0] = (sha->cnt[0] + n) & 0xffffffffffffffffULL;

	if (sha->cnt[0] < n) {
		sha->cnt[1] = (sha->cnt[1] + 1) & 0xffffffffffffffffULL;
	}
}

static
void sha512_blk_frm_buf (unsigned long long *W, const void *buf)
{
	unsigned i;

	for (i = 0; i < 16; i++) {
		W[i] = sha512_get_uint64 (buf, 8 * i);
	}

	for (i = 16; i < 80; i++) {
		W[i] = sha512_s1 (W[i - 2]) + W[i - 7] + sha512_s0 (W[i - 15]) + W[i - 16];
		W[i] &= 0xffffffffffffffffULL;
	}
}

static
void sha512_calc_block (sha512_t *sha, const void *buf)
{
	unsigned           i;
	unsigned long long h[8];
	unsigned long long W[80];
	unsigned long long t1, t2;

	sha512_blk_frm_buf (W, buf);

	for (i = 0; i < 8; i++) {
		h[i] = sha->H[i];
	}

	for (i = 0; i < 80; i++) {
		t1 = h[7] + sha512_S1 (h[4]) + sha512_F (h[4], h[5], h[6]) + sha512_K[i] + W[i];
		t2 = sha512_S0 (h[0]) + sha512_G (h[0], h[1], h[2]);

		h[7] = h[6];
		h[6] = h[5];
		h[5] = h[4];
		h[4] = sha512_add (h[3], t1);
		h[3] = h[2];
		h[2] = h[1];
		h[1] = h[0];
		h[0] = sha512_add (t1, t2);
	}

	for (i = 0; i < 8; i++) {
		sha->H[i] = sha512_add (sha->H[i], h[i]);
	}
}

void sha512_calc (sha512_t *sha, const void *buf, unsigned long cnt)
{
	unsigned            i, n;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	if (sha->buf_i > 0) {
		while ((sha->buf_i < 128) && (cnt > 0)) {
			sha->buf[sha->buf_i] = *tmp;
			tmp += 1;
			sha->buf_i += 1;
			cnt -= 1;
		}

		if (sha->buf_i < 128) {
			return;
		}

		sha512_calc_block (sha, sha->buf);

		sha->buf_i = 0;
		sha512_add_size (sha, 8 * 128);
	}

	while (cnt >= 128) {
		sha512_calc_block (sha, tmp);
		sha512_add_size (sha, 8 * 128);

		tmp += 128;
		cnt -= 128;
	}

	n = (unsigned) cnt;
	for (i = 0; i < n; i++) {
		sha->buf[i] = tmp[i];
	}

	sha->buf_i = n;
}

void sha512_done (sha512_t *sha)
{
	unsigned i;

	sha512_add_size (sha, 8 * sha->buf_i);

	sha->buf[sha->buf_i++] = 0x80;

	if (sha->buf_i > 112) {
		for (i = sha->buf_i; i < 128; i++) {
			sha->buf[i] = 0;
		}

		sha512_calc_block (sha, sha->buf);

		sha->buf_i = 0;
	}

	for (i = sha->buf_i; i < 112; i++) {
		sha->buf[i] = 0;
	}

	sha512_set_uint64 (sha->buf, 112, sha->cnt[1]);
	sha512_set_uint64 (sha->buf, 120, sha->cnt[0]);

	sha512_calc_block (sha, sha->buf);

	sha->buf_i = 0;
}

unsigned long long sha512_get_sum_i (sha512_t *sha, unsigned i)
{
	if (i < 8) {
		return (sha->H[i]);
	}

	return (0);
}

unsigned sha512_get_sum (sha512_t *sha, void *buf, unsigned max)
{
	unsigned i;

	if (max < 64) {
		return (0);
	}

	for (i = 0; i < 8; i++) {
		sha512_set_uint64 (buf, 8 * i, sha->H[i]);
	}

	return (64);
}

void sha512_calc1 (void *dst, const void *src, unsigned long n)
{
	sha512_t sha;

	sha512_init (&sha);
	sha512_calc (&sha, src, n);
	sha512_done (&sha);
	sha512_get_sum (&sha, dst, 64);
}
